@@ -6,7 +6,7 @@ use base qw[TestApp::View::Dump];
 sub process {
     my ( $self, $c ) = @_;
     my $r = $c->request;
-    local $r->{env};
+    #local $r->{env};
     return $self->SUPER::process( $c, $r );
 }
 
